/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.ow2.petals.tools.webconsole.util;

import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.management.ObjectName;
import javax.xml.namespace.QName;

import org.ow2.petals.jmx.api.IPetalsAdminService.ContainerKeyNames;
import org.ow2.petals.jmx.api.IPetalsAdminService.DomainKeyNames;
import org.ow2.petals.jmx.api.IPetalsAdminService.SubDomainKeyNames;
import org.ow2.petals.jmx.api.IPetalsAdminService.TopologyKeyNames;
import org.ow2.petals.jmx.exception.ComponentErrorException;
import org.ow2.petals.tools.webconsole.services.PetalsServiceException;
import org.ow2.petals.tools.webconsole.services.management.ManagementService;
import org.ow2.petals.tools.webconsole.services.management.ManagementServiceFactory;
import org.ow2.petals.tools.webconsole.to.ComponentTO;
import org.ow2.petals.tools.webconsole.to.DomainTO;
import org.ow2.petals.tools.webconsole.to.EndPointTO;
import org.ow2.petals.tools.webconsole.to.ServerTO;
import org.ow2.petals.tools.webconsole.to.ServiceAssemblyTO;

import static org.ow2.petals.jmx.api.IPetalsMonitoringService.ReportKeyNames.ENDPOINT_NAME;
import static org.ow2.petals.jmx.api.IPetalsMonitoringService.ReportKeyNames.SERVICE_NAME;
import static org.ow2.petals.jmx.api.IPetalsMonitoringService.ReportKeyNames.TYPE;

/**
 * @author aruffie
 *
 */
public class TopologyHelper {

    /**
     * Allow you to recover the current domain
     * with all servers into the specified domain
     * @param topology
     * @param host
     * @param port
     * @param login
     * @param password
     * @return
     * @throws PetalsServiceException 
     * @throws UnknownHostException 
     */
    public static DomainTO getDomain(final Set<Map<String, String>> topology,final String host,final Integer port,final String login,final String password ) throws PetalsServiceException, UnknownHostException {
        
        Iterator<Map<String,String>> entitiesIt = topology.iterator();
        String domainMode = null;
        /*
         *  Recover the global domain mode example:
         *  standalon, platorm, quickstart ...
         */
        while(entitiesIt.hasNext()){          
            Map<String,String> entitiesPropertiesMap = entitiesIt.next();
            if(entitiesPropertiesMap.containsKey(TopologyKeyNames.TOPOLOGY_TYPE)&&
                    entitiesPropertiesMap.get(TopologyKeyNames.TOPOLOGY_TYPE)
                    .equals(DomainKeyNames.TOPOLOGY_TYPE)){
                    domainMode = entitiesPropertiesMap
                                 .get(DomainKeyNames.CONF_DOMAIN_MODE);
            }
        }
        
        /*
         * recover all server into the topology
         */
        final Set<ServerTO> allServers = getAllServer(topology);
        /*
         * recover all domain into the topology
         */
        final Set<DomainTO> allDomains = getAllDomain(topology);
        
        /*
         * identify the good domain
         */
        final String domainName = identifyDomain(allServers,host,port,login,password);
        /*
         * find the good domain
         */
        final DomainTO currentDomain = findDomain(allDomains,domainName);
        if(currentDomain != null){
            currentDomain.setMode(domainMode);
            /*
             * recover all server for the good domain
             */
            currentDomain.getServers().addAll(recoverServer(allServers,domainName));
            
            /*
             * recover all entities for each servers
             */
            fillEntitiesForAllServer(currentDomain.getServers());
            return currentDomain;
        } else {
            throw new PetalsServiceException("Server: "+host+" not found in the current topology");
        }
    }

    /**
     * Allow to recover all servers present
     * in the specified domain
     * @param allServers
     * @param domainName
     * @return
     */
    private static Set<ServerTO> recoverServer(final Set<ServerTO> allServers,final String domainName) {
        Iterator<ServerTO> serversIt = allServers.iterator();
        final Set<ServerTO> goodServers = new HashSet<ServerTO>();
        while(serversIt.hasNext()){
            final ServerTO server = serversIt.next();
            if(server.getDomainName().equals(domainName)){
                goodServers.add(server);
            }
        }
        return goodServers;
    }

    /**
     * find the current domain into all
     * domains founded, by providing the
     * domain name you seek
     * @param allDomains
     * @param domainName
     * @return
     */
    private static DomainTO findDomain(final Set<DomainTO> allDomains,final String domainName) {
        
        Iterator<DomainTO> domainsIt = allDomains.iterator();
        while(domainsIt.hasNext()){
            final DomainTO domain = domainsIt.next();
            if(domain.getName().equals(domainName)){
                return domain;
            }
        }
        return null;
    }

    /**
     * Allow to identify the good domain
     * name for the current server, where
     * the connexion was etablished
     * @param allServers
     * @param host
     * @param port
     * @param login
     * @param password
     * @return
     * @throws UnknownHostException 
     */
    private static String identifyDomain(final Set<ServerTO> allServers, final String host,final Integer port,
            final String login,final String password) throws UnknownHostException {
        Iterator<ServerTO> serverIt = allServers.iterator();
        
       
        
        while(serverIt.hasNext()){
            final ServerTO server = serverIt.next();
            if(server.getHost().equals(host)&&
               server.getJmxPort().equals(port.toString())&&
               server.getJmxLogin().equals(login)&&
               server.getJmxPassword().equals(password)){
                return server.getDomainName();
            } else {
                server.setHost(GeneralHelper.replaceHostByIP(server.getHost()));
                if(server.getHost().equals(host)&&
                        server.getJmxPort().equals(port.toString())&&
                        server.getJmxLogin().equals(login)&&
                        server.getJmxPassword().equals(password)){
                         return server.getDomainName();
                }
            }
        }
        return null;
    }

    
    /**
     * Create a domain model object
     * by providing a map of its
     * properties
     * @param map
     * @return
     */
    private static DomainTO addDomain(final Map<String, String> map) {
        final DomainTO domain = new DomainTO();
        domain.setDescription(map.get(SubDomainKeyNames.CONF_SUBDOMAIN_DESCRIPTION));
        domain.setName(map.get(SubDomainKeyNames.CONF_SUBDOMAIN_NAME));
        domain.setNetworkSynchronized(Boolean.valueOf(map.get(SubDomainKeyNames.CONF_NETWORK_TIME_SYNCHRONIZED)));
        domain.setTopologyType(map.get(SubDomainKeyNames.TOPOLOGY_TYPE));
        return domain;
    }

    
    /**
     * Create a server model object
     * by providing a map of its
     * properties
     * @param serverPropertiesMap
     * @return
     */
    private static ServerTO addServer(final Map<String, String> serverPropertiesMap) {
        final ServerTO server = new ServerTO();
        server.setHost(serverPropertiesMap.get(ContainerKeyNames.CONF_HOST));
        server.setJmxLogin(serverPropertiesMap.get(ContainerKeyNames.CONF_USER));
        server.setJmxPort(serverPropertiesMap.get(ContainerKeyNames.CONF_JMX_RMI_PORT));
        server.setJmxPassword(serverPropertiesMap.get(ContainerKeyNames.CONF_PWD));
        server.setDescription(serverPropertiesMap.get(ContainerKeyNames.CONF_CONTAINER_DESCRIPTION));
        server.setTcpPort(serverPropertiesMap.get(ContainerKeyNames.CONF_TCP_PORT));
        server.setName(serverPropertiesMap.get(ContainerKeyNames.CONF_CONTAINER_NAME));
        server.setState(serverPropertiesMap.get(ContainerKeyNames.CONF_STATE));
        server.setNetworkBindPort(serverPropertiesMap.get(ContainerKeyNames.CONF_NETWORK_BIND_PORT));
        server.setDomainName(serverPropertiesMap.get(ContainerKeyNames.CONF_SUBDOMAIN_NAME));
        return server;     
    }
    
    /**
     * Allow to recover all servers into
     * all domains thank to a topology
     * set properties
     * @param topology
     * @return
     */
    private static Set<ServerTO> getAllServer(final Set<Map<String, String>> topology){
        final Set<ServerTO> servers = new HashSet<ServerTO>();
        Iterator<Map<String,String>> entitiesIt = topology.iterator();
        while(entitiesIt.hasNext()){
            Map<String,String> entitiesPropertiesMap = entitiesIt.next();
            
            /*
             * check if the current map is a container properties
             * map
             */
            if(entitiesPropertiesMap.containsKey(TopologyKeyNames.TOPOLOGY_TYPE)&&
               entitiesPropertiesMap.get(TopologyKeyNames.TOPOLOGY_TYPE)
               .equals(ContainerKeyNames.TOPOLOGY_TYPE)){
                servers.add(addServer(entitiesPropertiesMap));
            }
        }
        return servers;
    }
    
    /**
     * Allow to recover all domains into
     * a specific petals topology, by
     * providing a set properties
     * @param topology
     * @return
     */
    private static Set<DomainTO> getAllDomain(final Set<Map<String,String>> topology){
        Iterator<Map<String,String>> entitiesIt = topology.iterator();
        final Set<DomainTO> domains = new HashSet<DomainTO>();
        while(entitiesIt.hasNext()){
            Map<String,String> entitiesPropertiesMap = entitiesIt.next();
            if(entitiesPropertiesMap.containsKey(TopologyKeyNames.TOPOLOGY_TYPE)&&
                    entitiesPropertiesMap.get(TopologyKeyNames.TOPOLOGY_TYPE)
                    .equals(SubDomainKeyNames.TOPOLOGY_TYPE)){
                    domains.add(addDomain(entitiesPropertiesMap));
            }
        }
        return domains;
    }

    /**
     * Allow to recover all binding components
     * and their endpoints (for each binding
     * components) for the specified server
     * @param server
     * @param service
     * @throws PetalsServiceException
     */
    private static void recoverComponentsForServer(final ServerTO server,final ManagementService service) throws PetalsServiceException{
        ObjectName[] allComponents = service.componentsInventory();
        for(ObjectName objectName : allComponents){
            final ComponentTO component = new ComponentTO();
            component.setName(objectName.getKeyProperty("name"));
            try{
                component.setState(service.getComponentState(component.getName()));
            } catch (PetalsServiceException e){
                if(!(e.getCause() instanceof ComponentErrorException)){
                    throw e;
                } else {
                    component.setState("undefined");
                }
            }
            component.setType(ComponentTO.BC_TYPE);
            component.setEndpoints(recoverEndpointsForComponent(component.getName(), service, server.getName()));
            server.getBindingComponents().add(component);
        }
    }
    
    /**
     * Allow to recover all service engines
     * and their endpoints (for each service
     * engines) for the specified server
     * @param server
     * @param service
     * @throws PetalsServiceException
     */
    private static void recoverEnginesForServer(final ServerTO server,final ManagementService service) throws PetalsServiceException{
        ObjectName[] allEngines = service.enginesInventory();
        for(ObjectName objectName : allEngines){
            final ComponentTO component = new ComponentTO();
            component.setName(objectName.getKeyProperty("name"));
            try{
                component.setState(service.getComponentState(component.getName()));
            } catch (PetalsServiceException e){
                if(!(e.getCause() instanceof ComponentErrorException)){
                    throw e;
                } else {
                    component.setState("undefined");
                }
            }
            component.setType(ComponentTO.SE_TYPE);
            component.setEndpoints(recoverEndpointsForComponent(component.getName(), service,server.getName()));
            server.getServiceEngines().add(component);
        }
    }

    /**
     * Allow to recover all service assemblies
     * for a specific server
     * @param server
     * @param service
     * @throws PetalsServiceException
     */
    private static void recoverServiceAssembliesForServer(final ServerTO server,final ManagementService service) throws PetalsServiceException{
        for(final String sa : service.serviceAssembliesInventory()){
            final ServiceAssemblyTO serviceAssembly = new ServiceAssemblyTO();
            serviceAssembly.setName(sa);
            serviceAssembly.setState(service.getServiceAssemblyState(sa));
            server.getServiceAssemblies().add(serviceAssembly);
        }
    }
    
    /**
     * Allow to recover all shared libraries
     * for a specific server
     * @param server
     * @param service
     * @throws PetalsServiceException
     */
    private static void recoverSharedLibraries(final ServerTO server,final ManagementService service) throws PetalsServiceException{
        for(final String sharedLibrary : service.sharedLibrariesInventory()){
            server.getSharedLibraries().add(sharedLibrary);
        }
    }
    
    /**
     * Allow to recover all endpoint for a
     * specified component
     * @param componentName
     * @param service
     * @return
     * @throws PetalsServiceException
     */
    private static List<EndPointTO> recoverEndpointsForComponent(final String componentName,final ManagementService service, final String containerName) throws PetalsServiceException{
        final List<Map<String, Object>> allEndpoints = service.getAllEndpoints();
        final List<EndPointTO> endpointsTO = new ArrayList<EndPointTO>();
        for(final Map<String,Object> endpoint: allEndpoints){
            EndPointTO endpointTO = new EndPointTO();
            endpointTO.setComponent((String)endpoint.get("componentName"));
            endpointTO.setServer((String)endpoint.get("componentName"));
            endpointTO.setName((String)endpoint.get(ENDPOINT_NAME));
            endpointTO.setType((String)endpoint.get(TYPE));
            
            //Find the service by QName
            endpointTO.setService(QName.valueOf(
                                  (String)endpoint.get(SERVICE_NAME)
                                  ).toString());
            

            String interfaceNames = null;
            
            //begin identify all interfaces
            for(final String name : (String[])endpoint.get("interfaceNames")){
                if(StringHelper.isNullOrEmpty(interfaceNames)){
                    interfaceNames = QName.valueOf(name).toString();
                } else {
                    interfaceNames = interfaceNames +","+QName.valueOf(name);
                }
            }
            //Ending identify all interfaces
            
            endpointTO.setInterfaceName(interfaceNames);
            if(endpointTO.getComponent().equals(componentName) && (containerName.equals((String)endpoint.get("containerName")))){
                endpointsTO.add(endpointTO);
            }
        }
        return endpointsTO;
    }
    /**
     * Fill all entities for one server
     * binding components, services
     * engines, endpoints ...
     * @param server
     * @param service
     * @throws PetalsServiceException
     */
    public static void fillEntitiesForServer(final ServerTO server,final ManagementService service) throws PetalsServiceException{
        recoverComponentsForServer(server, service);
        recoverEnginesForServer(server, service);
        recoverServiceAssembliesForServer(server,service);
    }
    
    /**
     * Fill all entities for all servers
     * binding components, services
     * engines, endpoints ...
     * @param servers
     * @param service
     * @throws PetalsServiceException
     */
    public static void fillEntitiesForAllServer(final Collection<ServerTO> servers) throws PetalsServiceException{
        /* 
         * for each server etablish in first
         * the connexion to the service. 
         * Because all entities must be founded
         * for each specific server
         */
        boolean allServerUnAvailable = true;
        for(ServerTO server : servers){
            try{
                final ManagementService service = (ManagementService) 
                                              ManagementServiceFactory.getInstance()
                                              .getService(server.getHost(),
                                              Integer.valueOf(
                                              server.getJmxPort()),
                                              server.getJmxLogin(),
                                              server.getJmxPassword());
                recoverComponentsForServer(server, service);
                recoverEnginesForServer(server, service);
                recoverServiceAssembliesForServer(server, service);
                recoverSharedLibraries(server, service);
                allServerUnAvailable = false;
            } catch (final PetalsServiceException e){
                e.printStackTrace();
            }
        }
        
        if(allServerUnAvailable == true){
            throw new PetalsServiceException("All servers unavailable");
        }
    }

}
